#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# simple test where 
# the material anisotropy depends on temperature

import heat as h
import heat.utils as u
import tbox
import tbox.gmsh as gmsh
from fwk.testing import *

class MyMaterial(tbox.Fct2U):
    """
    Custom material
    calculates the conductivity tensor
    as a function of the local temperature
    """
    def __init__(self, kx, ky, theta):
        tbox.Fct2U.__init__(self)
        self.kx = kx
        self.ky = ky
        self.theta = theta
                            
    def eval(self, u, k, fake):
        import numpy as np
        import math
        t = (self.theta) *math.pi/180.0 
        u1 = 0. ; t1=-t
        u2 = 10. ; t2 = t
        
        t = t1+(t2-t1)/(u2-u1)*(u-u1)
        
        kx = self.kx
        ky = self.ky
        
        co = math.cos(t)
        si = math.sin(t)
        
        R = np.array([[co, -si],
                      [si, co]])
        Kt = np.array([[kx, 0.0],
                        [0.0, ky]])
        Tmp = np.dot(R,Kt)
        Kxy = np.dot(Tmp,R.T)
         
        k.resize(2, 2)
        k[0,0]=Kxy[0,0]
        k[1,1]=Kxy[1,1]
        k[1,0]=Kxy[1,0]
        k[0,1]=Kxy[0,1]
        #print k

def main():
    # mesh a unit square
    pars={ 'Lx' : 1 , 'Ly' : 1, 'Nx' : 20, 'Ny' : 10}
    msh = gmsh.MeshLoader("../../models/rect_stru.geo",__file__).execute(**pars)
    gmshWriter = tbox.GmshExport(msh)
    pbl = h.Problem(msh)

    # medium/boundary conditions
    myk = MyMaterial(1., 100., 20.)   # version python
    #myk=h.CompiledFct2b() # compiled version (not the same fct)
    pbl.add(h.Medium(msh, "S1", myk))
    pbl.add(h.Dirichlet(msh,"Left", 0.))
    pbl.add(h.Dirichlet(msh,"Right", 10.))

    # solve problem
    solver = u.solve(pbl, gmshWriter)

    # extract some results
    x1,y1,T1 = u.getTemp(solver,"Top", sort='x')
    
    u.simpleplot(x1, T1, xlabel='x [mm]', title='Temperature along x')
       
    tests = CTests()
    tests.add(CTest('mean(Ttop)', sum(T1) / len(T1), 5.0, 5e-2))
    tests.add(CTest('max(Ttop)', max(T1), 10, 1e-3))
    tests.add(CTest('min(Ttop)', min(T1), 0, 0.1))
    tests.run()


if __name__ == "__main__":
    main()
