/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTimeIntegration.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wQuad4.h"
#include "wTimer.h"
#include "wTag.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wBoundary.h"
#include "wWaveTerm.h"
#include "wFct0.h"
#include "wMshExport.h"
#include <typeinfo>
#include <deque>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>
using namespace fwk;
using namespace tbox;
using namespace waves;

TimeIntegration::TimeIntegration(std::shared_ptr<Problem> _pbl) : pbl(_pbl)
{
    dhook = NULL;

    // default values
    ttot = 10.0;
    dt = 1.0e-1 / 5;
    savefreq = 5;
    nthreads = 1;
    wgmsh = Wgmsh::WG_NONE;

    verbose = false;
    stopit = false;

    // Update element memory
    pbl->initElems();
}

void TimeIntegration::start(MshExport *mshWriter)
{
}

void TimeIntegration::dummyIC()
{
    std::cout << "initial condition...\n";

    auto msh = pbl->msh;

    // initial condition
    for (auto nod : msh->nodes)
    {
        double px = nod->pos(0);
        double py = nod->pos(1);
        double pz = nod->pos(2);
        double sx = 0.0;
        double sy = 0.0;
        double sz = 0.0;
        double R = sqrt((px - sx) * (px - sx) + (py - sy) * (py - sy) + (pz - sz) * (pz - sz));
        if (R < 1.0)
        {
            u1[nod->row] = 0.0;
            v1[nod->row] = 0.0;
        }
        else
        {
            u1[nod->row] = 0.0;
            v1[nod->row] = 0.0;
        }
    }
}

void TimeIntegration::buildS(Eigen::SparseMatrix<double, Eigen::RowMajor> &S2)
{
    tbb::spin_mutex mutex;
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    auto msh = pbl->msh;
    std::deque<Eigen::Triplet<double>> T; // list of triplets to build S matrix

    std::cout << "building S (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto bnd : pbl->bnds)
    {
        std::cout << "\tprocessing " << *bnd << '\n';
        std::cout << "normal=" << static_cast<Quad4 *>(bnd->tag->elems[0])->getNormal() << '\n';
        tbb::parallel_for_each(
            bnd->tag->elems.begin(), bnd->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::QUAD4)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                // ** Se matrix => S vector
                Eigen::MatrixXd Se = WaveTerm::buildM(*e);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, bnd->c * Se(ii, jj)));
                    }
                }
            });
    }
    // Build, clean and turn to compressed row format
    S2.setFromTriplets(T.begin(), T.end());
    S2.prune(0.);
    S2.makeCompressed();

    std::cout << "S (" << S2.rows() << "," << S2.cols() << ") nnz=" << S2.nonZeros() << "\n";
    chrono1.read();
    std::cout << "[cpu] " << chrono1 << '\n';
}

void TimeIntegration::buildKM(Eigen::SparseMatrix<double, Eigen::RowMajor> &K2,
                              std::vector<double> &Md, std::vector<double> const &u)
{
    Timer chrono;
    chrono.start();
    buildKM_tbb_lambda(K2, Md, u);
    chrono.stop();
    std::cout << "[cpu:buildKM] " << chrono << '\n';
}

void TimeIntegration::buildKM_tbb_lambda(Eigen::SparseMatrix<double, Eigen::RowMajor> &K2,
                                         std::vector<double> &Md, std::vector<double> const &u)
{
    tbb::spin_mutex mutex;
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    auto msh = pbl->msh;
    std::deque<Eigen::Triplet<double>> T; // list of triplets to build K matrix

    std::cout << "building K/M (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto mat : pbl->media)
    {
        std::cout << "\tprocessing " << *mat << '\n';
        tbb::parallel_for_each(
            mat->tag->elems.begin(), mat->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::HEX8)
                    return;
                // std::cout << "processing element #" << e->no << "\n";

                // ** Me matrix => Md vector
                Eigen::MatrixXd Me = WaveTerm::buildM(*e);

                // ** Ke matrix => K matrix
                Eigen::MatrixXd Ke = WaveTerm::buildK(*e, u);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, (mat->c * mat->c) * Ke(ii, jj)));
                        Md[nodi->row] += Me(ii, jj);
                    }
                }
            });
    }
    // Build, clean and turn to compressed row format
    K2.setFromTriplets(T.begin(), T.end());
    K2.prune(0.);
    K2.makeCompressed();

    std::cout << "S (" << K2.rows() << "," << K2.cols() << ") nnz=" << K2.nonZeros() << "\n";
    chrono1.read();
    std::cout << "[cpu] " << chrono1 << '\n';
}

/*
void TimeIntegration::build(MATTYPE type, Eigen::SparseMatrix<double, Eigen::RowMajor> &A2)
{
    tbb::spin_mutex mutex;
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads); // TODO mettre ça ailleurs...

    auto msh = pbl->msh;
    std::deque<Eigen::Triplet<double>> T; // list of triplets to build K matrix

    std::cout << "building " << type << " using " << nthreads << " threads...\n";

    for (auto mat : pbl->media)
    {
        std::cout << "\tprocessing " << *mat << '\n';
        tbb::parallel_for_each(
            mat->tag->elems.begin(), mat->tag->elems.end(),
            [&](Element *e)
            {
                if (e->type() != ElType::HEX8)
                    return;
                Eigen::MatrixXd Ae = e->build(type);

                // assembly
                tbb::spin_mutex::scoped_lock lock(mutex);
                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, Ae(ii, jj)));
                    }
                }
            });
    }
    // Build, clean and turn to compressed row format
    A2.setFromTriplets(T.begin(), T.end());
    A2.prune(0.);
    A2.makeCompressed();

    std::cout << "S (" << A2.rows() << "," << A2.cols() << ") nnz=" << A2.nonZeros() << "\n";
    chrono1.read();
    std::cout << "[cpu] " << chrono1 << '\n';
}
*/

void TimeIntegration::write(std::ostream &out) const
{
    out << clsname() << "\n";
    out << "\tttot     = " << ttot << "\n";
    out << "\tdt       = " << dt << "\n";
    out << "\tsavefreq = " << savefreq << "\n";
}
